using System;

namespace MyNamespace
{
	/// <summary>
	///   Calculation routines for FlyAKite program
	///   
	///   FlyAKite - A kite-flying program
	///   Developed in Apple ][ Pascal and converted to TURBO Pascal 4.0 10/30/88
	///   Apple ][ date 13-SEP-81, 1400 HRS
	///   Converted to TURBO Pascal 5.5 and 6.0, now in Visual C# .Net
	///   COPYRIGHT (c) 1981-2005 By author: Harry J Smith, Saratoga, CA
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	// ------- Parameters
	public class ParamCL
	{
		public int maxFlights;	// max # of flights per data file
		public int start;		// start line number for lines to compute wind factors
		public int stop;		// stop line number for lines to compute wind factors
		public int lineN;		// current Line number
		public int lines;		// number of stored lines
		public double h;		// integration step size in feet
		public double lOW;		// L/W, length of one pound of string in feet per pound
		public double fac1;		// low guess of wind factor
		public double fac2;		// high guess of wind factor

		public ParamCL()
		{
			;		// no constructor logic
		}
	}

	// ------- Flight data record 
	public class DataCL
	{
		public double length;	// length of string out in feet
		public double tension;	// tension in string at flyer in pounds
		public double zeta;		// angle to kite in degrees
		public double sA;		// angle of string at flyer in degrees
		public double factor;	// computed win factor
		public double xKite;	// computed x-coordinate of kite in feet
		public double yKite;	// computed y-coordinate of kite in feet

		public DataCL()
		{
			;		// no constructor logic
		}
	}

	// ------- Integration vector
	public class VectorCL
	{
		public double x;	// x-coordinate of kite
		public double y;	// y-coordinate of kite
		public double tX;	// tension string in the x direction
		public double tY;	// tension string in the y direction
		public double xD;	// x dot, dx/ds
		public double yD;	// y dot, dx/ds
		public double tXD;	// tx dot, d tx/ ds
		public double tYD;	// ty dot, d ty/ ds

		public VectorCL()
		{
			;		// no constructor logic
		}
	}
	
	// ------- Alternate form of integration vector
	public class Vector2CL
	{
		public double[] xA;		// array[1..4], xA[1] = x, xA[2] = y, xA[3] = tX, xA[4] = tY
		public double[] xDA;	// array[1..4], xDA[1] = xD, xDA[2] = yD, xDA[3] = tXD, xDA[4] = tYD

		public Vector2CL()
		{
			xA = new double[5];	// constructor logic
			xDA = new double[5];
		}
	}
	
	public class FlyAKite
	{
		public FlyAKite(int maxFlights)	// constructor logic
		{
			n = 4;						// compute constants
			nT = 100;
			eY = 0.0005;
			piO4 = Math.Atan(1.0);
			pi = 4.0 * piO4;
			pi2 = 8.0 * piO4;
			piO2 = 2.0 * piO4;
			dPR= 45.0 / piO4;
			Param = new ParamCL();				// initialize parameters, data records, and vertors
			Param.maxFlights = maxFlights;
			Data = new DataCL[maxFlights + 1];	// 1..maxFlights
			for (int i = 1; i <= maxFlights; i++)
			{
				Data[i] = new DataCL(); 
			}
			Vector = new VectorCL();
			Vector2 = new Vector2CL();
			NominalData(); // load one line of nominal data
		} // FlyAKite constructor logic

		public static void NominalData() // load one line of nominal data
		{
			Param.start = 1;
			Param.stop = Param.maxFlights;
			Param.lineN = 1;
			Param.lines = 1;
			Param.h = 600.0;
			Param.lOW = 8700.0;
			Param.fac1 = 20.0;
			Param.fac2 = 35.0;
			Data[1].length = 9315.0;
			Data[1].tension = 5.0;
			Data[1].zeta = 14.0;
			Data[1].sA = 3.0;
			Data[1].factor = 0.0;
			Data[1].xKite = 0.0;
			Data[1].yKite = 0.0;
		}

		public static void OutputIt() // output iteration results
		{
			double r;
			r = Math.Sqrt(Vector.x * Vector.x + Vector.y * Vector.y);
			zetaC = dPR * Math.Atan2(Vector.y, Vector.x);
			sAC = dPR * Math.Atan2(Vector.tY, Vector.tX);
			string st = "";
			st += string.Format("{0,8:0.0}", s);
			st += string.Format(" {0,7:0.0}", Vector.x);
			st += string.Format(" {0,7:0.0}", Vector.y);
			st += string.Format(" {0,7:0.00}", Vector.tX);
		 	st += string.Format(" {0,7:0.00}", Vector.tY);
			st += string.Format(" {0,7:0.00}", tensionC);
			st += string.Format(" {0,7:0.000}", sAC);
			st += string.Format(" {0,7:0.000}", zetaC);
			st += string.Format(" {0,7:0.0}", r);
			st += string.Format(" {0,7:0.00}", factorC);
			Common.WriteLn(st);
		}

		public static void RHSF() // Evaluate the Right Hand Side Function of ordinary differential equation
		// the four variables (x, y, tX, tY) are a function of s:
		// dx/ds = tX / t
		// dy/ds = tY / t
		// d(tX)/ds = -eMu * Abs((dy/ds)^3)
		// d(ty)/ds = wOL + eMu * Abs(dy/ds) * dy/ds * dx/ds
		// where t = Sqrt(tX^2 + tY^2)
		// eMu = wOL * factor, a constant
		// wOL = W/L, weight of one foot of string in pounds per foot, a constant
		// factor = wind factor, a constant
		{
			tensionC = Math.Sqrt(Vector.tX * Vector.tX + Vector.tY * Vector.tY);
			Vector.xD = Vector.tX / tensionC;
			Vector.yD = Vector.tY / tensionC;
			Vector.tXD = -eMu * Math.Abs(Vector.yD * Vector.yD * Vector.yD);
			Vector.tYD = wOL + eMu * Math.Abs(Vector.yD) * Vector.yD * Vector.xD;
		}

		public static void RHSF2() // Evaluate the Right Hand Side Function using Vectot2
		{
			tensionC = Math.Sqrt(Vector2.xA[3] * Vector2.xA[3] + Vector2.xA[4] * Vector2.xA[4]);
			Vector2.xDA[1] = Vector2.xA[3] / tensionC;
			Vector2.xDA[2] = Vector2.xA[4] / tensionC;
			Vector2.xDA[3] = -eMu * Math.Abs(Vector2.xDA[2] * Vector2.xDA[2] * Vector2.xDA[2]);
			Vector2.xDA[4] = wOL + eMu * Math.Abs(Vector2.xDA[2]) * Vector2.xDA[2] * Vector2.xDA[1];
		}

		public static void RKutta(ref double s, double h)	// must call RHSF before RKutta
		// Runge-Kutta method of integration of ordinary differential equations
		{
			Vector2.xA[1] = Vector.x;	// copy Vector to Vector2
			Vector2.xA[2] = Vector.y;
			Vector2.xA[3] = Vector.tX;
			Vector2.xA[4] = Vector.tY;
			Vector2.xDA[1] = Vector.xD;
			Vector2.xDA[2] = Vector.yD;
			Vector2.xDA[3] = Vector.tXD;
			Vector2.xDA[4] = Vector.tYD;
			double[] eK1;
			double[] eK2;
			double[] eK3;
			double[] xI;
			double sR;
			int i;
			eK1 = new double[n+1];
			eK2 = new double[n+1];
			eK3 = new double[n+1];
			xI = new double[n+1];
			sR = s;
			for (i = 1; i <= n; i++)
			{
				xI[i] = Vector2.xA[i];
				eK1[i] = h * Vector2.xDA[i];
				Vector2.xA[i] = xI[i] + eK1[i] / 2.0;
			}
			s = sR + h / 2.0;
			RHSF2();
			for (i = 1; i <= n; i++)
			{
				eK2[i] = h * Vector2.xDA[i];
				Vector2.xA[i] = xI[i] + eK2[i] / 2.0;
			}
			RHSF2();
			for (i = 1; i <= n; i++)
			{
				eK3[i] = h * Vector2.xDA[i];
				Vector2.xA[i] = xI[i] + eK3[i];
			}
			s = sR + h;
			RHSF2();
			for (i = 1; i <= n; i++)
			{
				Vector2.xA[i] = xI[i] +
					(eK1[i] + 2.0 * eK2[i] + 2.0 * eK3[i] + h * Vector2.xDA[i]) / 6.0;
			}
			Vector.x = Vector2.xA[1];	// copy Vector2 to Vector
			Vector.y = Vector2.xA[2];
			Vector.tX = Vector2.xA[3];
			Vector.tY = Vector2.xA[4];
			Vector.xD = Vector2.xDA[1];
			Vector.yD = Vector2.xDA[2];
			Vector.tXD = Vector2.xDA[3];
			Vector.tYD = Vector2.xDA[4];
		} // RKutta

		public static void AIRK(ref double s, double length, double h)	// Advance Integration by RKutta
		// integrate from s to length with a step size of h
		{
			if (length == s) return;
			h = Math.Abs(h);
			if (length < s) h = -h;
			while (Math.Abs(length - s) > Math.Abs(h))
			{
				RKutta(ref s, h);
				RHSF();
			}
			h = length - s;
			if (h != 0) RKutta(ref s, h);
			s = length;
		} // AIRK

		public static void ResetToFlyer()	// reset conditions to the flyer
		{
			eMu = wOL * factorC;
			s = 0;					// reset conditions to the flyer
			Vector.x = 0;
			Vector.y = 0;
			Vector.tX = tXIn;
			Vector.tY = tYIn;
			RHSF();
		}

		public static double Function(double x) // find the root of this function
		// error in computed value of zeta as a function of wind factor
		{
			factorC = x;
			ResetToFlyer();				// reset conditions to the flyer
			AIRK(ref s, lengthD, hP);
			RHSF();
			OutputIt();
			return zetaC - zetaD;
		}

		public static double Root(double x1, double x2, double eY, int nT)
		// find root of y = Function(x). x1 and x2 are the first two guesses, hopefully bracketing
		// the root. eY is the allowed error in Abs(y), nT is the maximum number of iterations to try.
		// the value returned is the root x. converge and abort flags are also set
		{
//			eY = 0;					// for test only
//			nT = Int32.MaxValue;	// for test only
			double x;
			double y;
			double y1;
			double y2;
			double p;
			int i;
			converge = false;
			y2 = Function(x2);
			x = x1;
			y = Function(x);
			for (i = 1; i <= nT; i++)
			{
				if (Math.Abs(y) > Math.Abs(y2))
				{
					x1 = x;
					y1 = y;
				}
				else
				{
					x1 = x2;
					y1 = y2;
					x2 = x;
					y2 = y;
				}
				if (y1 != y2)
				{
					p = y2 / (y1 - y2);
					if (p > 1.0) p = 1.0;
				}
				else
				{
					p = 1.0;
				}
				x = x2 + p * (x2 - x1);
				y = Function(x);
				if (Math.Abs(y) <= eY)
				{
					Common.WriteLn("Converged");
					if (Global.objRunForm.buttonPlot.CanFocus) // test CanFocus
						Global.objRunForm.buttonPlot.Focus(); // move from output text box to buttonPlot
					converge = true;
					return x;
				}
				Common.MuInterrupt();
				if (Common.abortCalc)
				{
					Common.WriteLn("Manually aborted before convergence");
					return 0.0;
				}
			}
			Common.WriteLn("Did not converge in " + nT + " iterations");
			return 0.0;
		} // Root

		public static void GetParam()
		{
			fac1P = Param.fac1;
			fac2P = Param.fac2;
			lOWP = Param.lOW;
			hP = Param.h;
			if (lOWP <= 0) return;
			wOL = 1.0 / lOWP;
		} // GetParam

		public static void GetCase(int i)
		{
			lengthD = Data[i].length;
			tensionD = Data[i].tension;
			zetaD = Data[i].zeta;
			sAD = Data[i].sA;
			factorC = Data[i].factor;
			tYIn = tensionD * Math.Sin(sAD / dPR);
			tXIn = tensionD * Math.Cos(sAD / dPR);
			s = 0;
			Vector.x = 0;
			Vector.y = 0;
			Vector.tX = tXIn;
			Vector.tY = tYIn;
			if (lOWP <= 0) return;
			RHSF();
		} // GetCase

		public static void CompFactor() // Compute wind factor
		{
			if (Common.running) return;			// don't call again till its returns
			Common.running = true;
			Common.abortCalc = false;
			int start = Param.start;
			int stop = Param.stop;
			if (start > Param.lines) start = Param.lines;
			if (stop > Param.lines) stop = Param.lines;
			if (stop < start) stop = start;
			Common.WriteDLn("Calculating wind factors for flights " + start + " through " + stop);
			GetParam();
			for (int i = start; i <= stop; i++)
			{
				if ((lOWP <= 0) || Common.abortCalc) break;
				GetCase(i);
				if ((lOWP > 0) && (tensionC > 0))
				{
					Common.WriteDLn("   Len       T       Zeta      SA     Fac1    Fac2    L/W      H      Flight");
					string st = "";
					st += string.Format("{0,8:0.0}", lengthD);
					st += string.Format(" {0,7:0.00}", tensionD);
					st += string.Format(" {0,9:0.000}", zetaD);
					st += string.Format(" {0,8:0.000}", sAD);
					st += string.Format(" {0,7:0.00}", fac1P);
					st += string.Format(" {0,7:0.00}", fac2P);
					st += string.Format(" {0,7:0.0}", lOWP);
					st += string.Format(" {0,7:0.0}", hP);
					st += string.Format(" {0,6:0}", i);
					Common.WriteLn(st);

					Common.WriteDLn("   Len      X       Y       TX      TY       T      SA     Zeta" +
					"     R     Factor");
					factorC = Root(fac1P, fac2P, eY, nT);
					if (converge)
					{
						Data[i].xKite = Vector.x;
						Data[i].yKite = Vector.y;
						Data[i].factor = factorC;
					}
					else
					{
						Data[i].xKite = 0;
						Data[i].yKite = 0;
						Data[i].factor = 0;
					}
				}
			}
			Common.running = false;
			Common.abortCalc = false;
		} // CompFactor
		
		// Global parameters
		public static ParamCL Param;	// collection of parameters
		public static double fac1P;		// low guess for factor
		public static double fac2P;		// high guess for factor
		public static double hP;		// integration step size in feet
		public static double lOWP;		// L/W, length of one pound of string in feet per pound

		// Global data
		public static DataCL[] Data;	// array of all flight data
		public static double lengthD;	// length of string out in feet
		public static double tensionD;	// tension in string at flyer in pounds
		public static double sAD;		// angle of string at flyer in degrees
		public static double zetaD;		// angle to kite in degrees

		// Global constants
		public static double dPR;		// degrees per radian
		public static double eY;		// epsilon for accuracy of computed zeta in degrees
		public static double pi;		// pi = 3.14159...
		public static double pi2;		// 2 * pi
		public static double piO2;		// pi / 2
		public static double piO4;		// pi / 4
		public static int n;			// n = 4 for 4th order Runge-Kutta method
		public static int nT;			// Maximum number of iterations for convergence
		
		// Global variables
		public static bool converge;	// converge flag
		public static double eMu;		// eMu = wOL * factorD
		public static double factorC;	// computed wind factor
		public static double s;			// independent variable s, feet of string from flyer
		public static double sAC;		// computed angle of string in degrees
		public static double tensionC;	// tension in the string at s feet of string from flyer
		public static double tXIn;		// x component of tension in string at flyer in pounds
		public static double tYIn;		// y component of tension in string at flyer in pounds
		public static double wOL;		// W/L, weight of one foot of string in pounds per foot
		public static double zetaC;		// computed angle to kite in degrees
		public static VectorCL Vector;	// vector
		public static Vector2CL Vector2;// alternate form of vector
	}
} // End FlyAKite
